# encoding=utf-8

# This version of this file is only used when run uninstalled. It is replaced
# with a generated version during installation.
from game_data_packager.version import (
        GAME_PACKAGE_RELEASE,
        GAME_PACKAGE_VERSION,
        )

# reassure pyflakes
GAME_PACKAGE_RELEASE
GAME_PACKAGE_VERSION
